@@ -1,5 +1,6 @@
+from __future__ import annotations
+
 import enum
-import typing
 import logging
 import dataclasses
 
@@ -10,15 +11,6 @@ import zigpy_znp.types as t
 LOGGER = logging.getLogger(__name__)
 
 
-class BindEntry(t.Struct):
-    """Bind table entry."""
-
-    Src: t.EUI64
-    SrcEp: t.uint8_t
-    ClusterId: t.ClusterId
-    DstAddr: zigpy.zdo.types.MultiAddress
-
-
 class CommandType(t.enum_uint8):
     """Command Type."""
 
@@ -160,8 +152,8 @@ class CommandHeader(t.uint16_t):
 class CommandDef:
     command_type: CommandType
     command_id: t.uint8_t
-    req_schema: typing.Optional[tuple] = None
-    rsp_schema: typing.Optional[tuple] = None
+    req_schema: tuple | None = None
+    rsp_schema: tuple | None = None
 
 
 class CommandsMeta(type):
@@ -384,21 +376,28 @@ class CommandBase:
 
         super().__setattr__("_bound_params", bound_params)
 
-    def to_frame(self):
+    def to_frame(self, *, align=False):
         if self._partial:
             raise ValueError(f"Cannot serialize a partial frame: {self}")
 
         from zigpy_znp.frames import GeneralFrame
 
-        # At this point the optional params are assumed to be in a valid order
-        data = b"".join(
-            [v.serialize() for p, v in self._bound_params.values() if v is not None]
-        )
+        chunks = []
 
-        return GeneralFrame(self.header, data)
+        for param, value in self._bound_params.values():
+            # At this point the optional params are assumed to be in a valid order
+            if value is None:
+                continue
+
+            if issubclass(param.type, t.CStruct):
+                chunks.append(value.serialize(align=align))
+            else:
+                chunks.append(value.serialize())
+
+        return GeneralFrame(self.header, b"".join(chunks))
 
     @classmethod
-    def from_frame(cls, frame, *, ignore_unparsed=False) -> "CommandBase":
+    def from_frame(cls, frame, *, align=False) -> "CommandBase":
         if frame.header != cls.header:
             raise ValueError(
                 f"Wrong frame header in {cls}: {cls.header} != {frame.header}"
@@ -409,7 +408,10 @@ class CommandBase:
 
         for param in cls.schema:
             try:
-                params[param.name], data = param.type.deserialize(data)
+                if issubclass(param.type, t.CStruct):
+                    params[param.name], data = param.type.deserialize(data, align=align)
+                else:
+                    params[param.name], data = param.type.deserialize(data)
             except ValueError:
                 if not data and param.optional:
                     # If we're out of data and the parameter is optional, we're done
@@ -425,16 +427,10 @@ class CommandBase:
                     raise
 
         if data:
-            msg = (
-                f"Unparsed data remains at the end of the frame while parsing {cls}"
-                f" (parsed {params}): {data!r}"
+            raise ValueError(
+                f"Frame {frame} contains trailing data after parsing: {data}"
             )
 
-            if ignore_unparsed:
-                LOGGER.warning(msg)
-            else:
-                raise ValueError(msg)
-
         return cls(**params)
 
     def matches(self, other: "CommandBase") -> bool:
@@ -547,25 +543,25 @@ class InterPanCommand(t.enum_uint8):
 
 
 class MTCapabilities(t.enum_flag_uint16):
-    CAP_SYS = 1 << 0
-    CAP_MAC = 1 << 1
-    CAP_NWK = 1 << 2
-    CAP_AF = 1 << 3
-    CAP_ZDO = 1 << 4
-    CAP_SAPI = 1 << 5
-    CAP_UTIL = 1 << 6
-    CAP_DEBUG = 1 << 7
-    CAP_APP = 1 << 8
-    CAP_GP = 1 << 9
-    CAP_APP_CNF = 1 << 10
-    CAP_UNK12 = 1 << 11
-    CAP_ZOAD = 1 << 12
-    CAP_UNK14 = 1 << 13
-    CAP_UNK15 = 1 << 14
-    CAP_UNK16 = 1 << 15
+    SYS = 1 << 0
+    MAC = 1 << 1
+    NWK = 1 << 2
+    AF = 1 << 3
+    ZDO = 1 << 4
+    SAPI = 1 << 5
+    UTIL = 1 << 6
+    DEBUG = 1 << 7
+    APP = 1 << 8
+    GP = 1 << 9
+    APP_CNF = 1 << 10
+    UNK12 = 1 << 11
+    ZOAD = 1 << 12
+    UNK14 = 1 << 13
+    UNK15 = 1 << 14
+    UNK16 = 1 << 15
 
 
-class Network(t.Struct):
+class Network(t.CStruct):
     PanId: t.PanId
     Channel: t.uint8_t
     StackProfileVersion: t.uint8_t
