@@ -26,6 +26,7 @@ import zigpy_znp.types as t
 import zigpy_znp.config as conf
 import zigpy_znp.commands as c
 from zigpy_znp.api import ZNP
+from zigpy_znp.utils import combine_concurrent_calls
 from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
 from zigpy_znp.types.nvids import OsalNvIds
 from zigpy_znp.zigbee.zdo_converters import ZDO_CONVERTERS
@@ -44,7 +45,7 @@ BROADCAST_SEND_WAIT_DURATION = 3
 MULTICAST_SEND_WAIT_DURATION = 3
 
 REQUEST_MAX_RETRIES = 5
-REQUEST_ERROR_RETRY_DELAY = 0.5  # seconds
+REQUEST_ERROR_RETRY_DELAY = 0.5
 
 # Errors that go away on their own after waiting for a bit
 REQUEST_TRANSIENT_ERRORS = {
@@ -91,17 +92,16 @@ class ZNPCoordinator(zigpy.device.Device):
 
     @property
     def model(self):
-        if self.application._znp.version == 1.2:
-            model = "CC2531"
-            version = "Home 1.2"
-        elif self.application._znp.version == 3.0:
-            model = "CC2531"
-            version = "3.0.1/3.0.2"
+        if self.application._znp.version > 3.0:
+            model = "CC1352/CC2652"
+            version = "3.30+"
         else:
-            model = "CC13X2/CC26X2"
-            version = "3.30.00/3.40.00/4.10.00"
+            model = "CC2538" if self.application._znp.nvram.align_structs else "CC2531"
+            version = "Home 1.2" if self.application._znp.version == 1.2 else "3.0.x"
 
-        return f"{model}, Z-Stack {version}"
+        build = self.application._version_rsp.CodeRevision
+
+        return f"{model}, Z-Stack {version} (build {build})"
 
 
 class ControllerApplication(zigpy.application.ControllerApplication):
@@ -122,9 +122,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         self._network_key = None
         self._network_key_seq = None
+        self._version_rsp = None
         self._concurrent_requests_semaphore = None
         self._currently_waiting_requests = 0
-        self._route_discovery_futures = {}
+
         self._join_announce_tasks = {}
 
     ##################################################################
@@ -175,11 +176,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self._reconnect_task.cancel()
         self._watchdog_task.cancel()
 
-        for f in self._route_discovery_futures.values():
-            f.cancel()
-
-        self._route_discovery_futures.clear()
-
         # This will close the UART, which will then close the transport
         if self._znp is not None:
             self._znp.close()
@@ -258,11 +254,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # At this point the device state should the same, regardless of whether we just
         # formed a new network or are restoring one
         if self.znp_config[conf.CONF_TX_POWER] is not None:
-            dbm = self.znp_config[conf.CONF_TX_POWER]
-
-            await self._znp.request(
-                c.SYS.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.SUCCESS
-            )
+            await self.set_tx_power(dbm=self.znp_config[conf.CONF_TX_POWER])
 
         # Both versions of Z-Stack use this callback
         started_as_coordinator = self._znp.wait_for_response(
@@ -303,6 +295,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         async with async_timeout.timeout(STARTUP_TIMEOUT):
             await started_as_coordinator
 
+        self._version_rsp = await self._znp.request(c.SYS.Version.Req())
+
         # XXX: The CC2531 running Z-Stack Home 1.2 permanently permits joins on startup
         # unless they are explicitly disabled. We can't fix this but we can disable them
         # as early as possible to shrink the window of opportunity for unwanted joins.
@@ -314,36 +308,11 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             await self._set_led_mode(led=0xFF, mode=self.znp_config[conf.CONF_LED_MODE])
 
         await self._load_network_info()
+        await self._register_endpoints()
 
-        # Add the coordinator as a zigpy device. We do this up here because
-        # `self._register_endpoint()` adds endpoints to this device object.
+        # Setup the coordinator as a zigpy device and initialize it to request node info
         self.devices[self.ieee] = ZNPCoordinator(self, self.ieee, self.nwk)
-
-        # Give our Zigpy device a valid node descriptor
-        node_descriptor_rsp = await self._znp.request_callback_rsp(
-            request=c.ZDO.NodeDescReq.Req(DstAddr=0x0000, NWKAddrOfInterest=0x0000),
-            RspStatus=t.Status.SUCCESS,
-            callback=c.ZDO.NodeDescRsp.Callback(Src=0x0000, NWK=0x0000, partial=True),
-        )
-        self.zigpy_device.node_desc = node_descriptor_rsp.NodeDescriptor
-
-        # Register our endpoints
-        await self._register_endpoint(
-            endpoint=1,
-            profile_id=zigpy.profiles.zha.PROFILE_ID,
-            device_id=zigpy.profiles.zha.DeviceType.IAS_CONTROL,
-            input_clusters=[clusters.general.Ota.cluster_id],
-            output_clusters=[
-                clusters.security.IasZone.cluster_id,
-                clusters.security.IasWd.cluster_id,
-            ],
-        )
-
-        await self._register_endpoint(
-            endpoint=2,
-            profile_id=zigpy.profiles.zll.PROFILE_ID,
-            device_id=zigpy.profiles.zll.DeviceType.CONTROLLER,
-        )
+        await self.zigpy_device.schedule_initialize()
 
         # Now that we know what device we are, set the max concurrent requests
         if self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS] == "auto":
@@ -353,11 +322,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         self._concurrent_requests_semaphore = asyncio.Semaphore(max_concurrent_requests)
 
-        version_rsp = await self._znp.request(c.SYS.Version.Req())
-
         LOGGER.info("Network settings")
+        LOGGER.info("  Model: %s", self.zigpy_device.model)
         LOGGER.info("  Z-Stack version: %s", self._znp.version)
-        LOGGER.info("  Z-Stack build id: %s", version_rsp.CodeRevision)
+        LOGGER.info("  Z-Stack build id: %s", self._version_rsp.CodeRevision)
         LOGGER.info("  Max concurrent requests: %s", max_concurrent_requests)
         LOGGER.info("  Channel: %s", self.channel)
         LOGGER.info("  PAN ID: 0x%04X", self.pan_id)
@@ -416,6 +384,25 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._znp.nvram.osal_write(OsalNvIds.NIB, nib)
         await self._znp.nvram.osal_write(OsalNvIds.PANID, pan_id)
 
+    async def set_tx_power(self, dbm: int) -> None:
+        """
+        Sets the radio TX power.
+        """
+
+        rsp = await self._znp.request(c.SYS.SetTxPower.Req(TXPower=dbm))
+
+        if self._znp.version >= 3.30 and rsp.StatusOrPower != t.Status.SUCCESS:
+            # Z-Stack 3's response indicates success or failure
+            raise InvalidCommandResponse(
+                f"Failed to set TX power: {t.Status(rsp.StatusOrPower)!r}", rsp
+            )
+        elif self._znp.version < 3.30 and rsp.StatusOrPower != dbm:
+            # Old Z-Stack releases used the response status field to indicate the power
+            # setting that was actually applied
+            LOGGER.warning(
+                "Requested TX power %d was adjusted to %d", dbm, rsp.StatusOrPower
+            )
+
     async def form_network(self):
         """
         Clears the current config and forms a new network with a random network key,
@@ -710,19 +697,24 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         Permit joining the network via a specific node or via all router nodes.
         """
 
-        # Always permit joins on the coordinator first.
-        # This unfortunately makes it impossible to permit joins via just one router
-        # alone but firmware changes can make this possible on newer hardware.
-        response = await self._znp.request_callback_rsp(
-            request=c.ZDO.MgmtPermitJoinReq.Req(
-                AddrMode=t.AddrMode.NWK,
-                Dst=0x0000,
-                Duration=time_s,
-                TCSignificance=1,
-            ),
-            RspStatus=t.Status.SUCCESS,
-            callback=c.ZDO.MgmtPermitJoinRsp.Callback(Src=0x0000, partial=True),
-        )
+        LOGGER.info("Permitting joins for %d seconds", time_s)
+
+        if time_s == 0 or self._version_rsp.CodeRevision < 20210708:
+            # If joins were permitted through a specific router, older Z-Stack builds
+            # did not allow the key to be distributed unless the coordinator itself was
+            # also permitting joins.
+            #
+            # Fixed in https://github.com/Koenkk/Z-Stack-firmware/commit/efac5ee46b9b437
+            response = await self._znp.request_callback_rsp(
+                request=c.ZDO.MgmtPermitJoinReq.Req(
+                    AddrMode=t.AddrMode.NWK,
+                    Dst=0x0000,
+                    Duration=time_s,
+                    TCSignificance=1,
+                ),
+                RspStatus=t.Status.SUCCESS,
+                callback=c.ZDO.MgmtPermitJoinRsp.Callback(Src=0x0000, partial=True),
+            )
 
         if response.Status != t.Status.SUCCESS:
             raise RuntimeError(f"Failed to permit joins on the coordinator: {response}")
@@ -830,6 +822,30 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             c.ZDO.PermitJoinInd.Callback(partial=True), self.on_zdo_permit_join_message
         )
 
+        # No-op handle commands that just create unnecessary WARNING logs
+        self._znp.callback_for_response(
+            c.ZDO.ParentAnnceRsp.Callback(partial=True),
+            self.on_intentionally_unhandled_message,
+        )
+
+        self._znp.callback_for_response(
+            c.ZDO.ConcentratorInd.Callback(partial=True),
+            self.on_intentionally_unhandled_message,
+        )
+
+        self._znp.callback_for_response(
+            c.ZDO.MgmtNWKUpdateNotify.Callback(partial=True),
+            self.on_intentionally_unhandled_message,
+        )
+
+    def on_intentionally_unhandled_message(self, msg: t.CommandBase) -> None:
+        """
+        Some commands are unhandled but frequently sent by devices on the network. To
+        reduce unnecessary logging messages, they are given an explicit callback.
+        """
+
+        pass
+
     def on_zdo_permit_join_message(self, msg: c.ZDO.PermitJoinInd.Callback) -> None:
         """
         Coordinator join status change message. Only sent with Z-Stack 1.2 and 3.0.
@@ -840,18 +856,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         else:
             LOGGER.info("Coordinator is permitting joins for %d seconds", msg.Duration)
 
-    def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
+    async def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
         """
         ZDO source routing message callback
         """
 
-        LOGGER.info("ZDO device relays: %s", msg)
+        device = await self._get_or_discover_device(nwk=msg.DstAddr)
 
-        try:
-            device = self.get_device(nwk=msg.DstAddr)
-        except KeyError:
+        if device is None:
             LOGGER.warning(
-                "Received a ZDO message from an unknown device: 0x%04x", msg.DstAddr
+                "Received a ZDO message from an unknown device: %s", msg.DstAddr
             )
             return
 
@@ -870,11 +884,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             self._join_announce_tasks.pop(msg.IEEE).cancel()
 
         # Sometimes devices change their NWK when announcing so re-join it.
-        self.handle_join(
-            nwk=msg.NWK,
-            ieee=msg.IEEE,
-            parent_nwk=None,
-        )
+        self.handle_join(nwk=msg.NWK, ieee=msg.IEEE, parent_nwk=None)
 
         device = self.get_device(ieee=msg.IEEE)
 
@@ -902,6 +912,19 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if msg.SrcIEEE in self._join_announce_tasks:
             self._join_announce_tasks.pop(msg.SrcIEEE).cancel()
 
+        # If the device already exists, immediately trigger a join to update its NWK.
+        try:
+            self.get_device(ieee=msg.SrcIEEE)
+        except KeyError:
+            pass
+        else:
+            self.handle_join(
+                nwk=msg.SrcNwk,
+                ieee=msg.SrcIEEE,
+                parent_nwk=msg.ParentNwk,
+            )
+            return
+
         # Some devices really don't like zigpy beginning its initialization process
         # before the device has announced itself. Wait a second or two before calling
         # `handle_join`, just in case the device announces itself first.
@@ -918,16 +941,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         LOGGER.info("ZDO device left: %s", msg)
         self.handle_leave(nwk=msg.NWK, ieee=msg.IEEE)
 
-    def on_af_message(self, msg: c.AF.IncomingMsg.Callback) -> None:
+    async def on_af_message(self, msg: c.AF.IncomingMsg.Callback) -> None:
         """
         Handler for all non-ZDO messages.
         """
 
-        try:
-            device = self.get_device(nwk=msg.SrcAddr)
-        except KeyError:
+        device = await self._get_or_discover_device(nwk=msg.SrcAddr)
+
+        if device is None:
             LOGGER.warning(
-                "Received an AF message from an unknown device: 0x%04x", msg.SrcAddr
+                "Received an AF message from an unknown device: %s", msg.SrcAddr
             )
             return
 
@@ -996,7 +1019,69 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
                 return
 
-    async def _set_led_mode(self, *, led, mode) -> None:
+    @combine_concurrent_calls
+    async def _get_or_discover_device(self, nwk: t.NWK) -> zigpy.device.Device | None:
+        """
+        Finds a device by its NWK address. If a device does not exist in the zigpy
+        database, attempt to look up its new NWK address. If it does not exist in the
+        zigpy database, treat the device as a new join.
+        """
+
+        try:
+            return self.get_device(nwk=nwk)
+        except KeyError:
+            pass
+
+        LOGGER.debug("Device with NWK 0x%04X not in database", nwk)
+
+        try:
+            # XXX: Multiple responses may arrive but we only use the first one
+            ieee_addr_rsp = await self._znp.request_callback_rsp(
+                request=c.ZDO.IEEEAddrReq.Req(
+                    NWK=nwk,
+                    RequestType=c.zdo.AddrRequestType.SINGLE,
+                    StartIndex=0,
+                ),
+                RspStatus=t.Status.SUCCESS,
+                callback=c.ZDO.IEEEAddrRsp.Callback(
+                    partial=True,
+                    NWK=nwk,
+                ),
+                timeout=5,  # We don't want to wait forever
+            )
+        except asyncio.TimeoutError:
+            return None
+
+        ieee = ieee_addr_rsp.IEEE
+
+        try:
+            device = self.get_device(ieee=ieee)
+        except KeyError:
+            LOGGER.debug("Treating unknown device as a new join")
+            self.handle_join(nwk=nwk, ieee=ieee, parent_nwk=None)
+
+            return self.get_device(ieee=ieee)
+
+        # The `Device` object could have been updated while this coroutine is running
+        if device.nwk == nwk:
+            return device
+
+        LOGGER.warning(
+            "Device %s changed its NWK from %s to %s",
+            device.ieee,
+            device.nwk,
+            nwk,
+        )
+
+        # Notify zigpy of the change
+        self.handle_join(nwk=nwk, ieee=ieee, parent_nwk=None)
+
+        # `handle_join` will update the NWK
+        assert device.nwk == nwk
+
+        return device
+
+    async def _set_led_mode(self, *, led: t.uint8_t, mode: c.util.LEDMode) -> None:
         """
         Attempts to set the provided LED's mode. A Z-Stack bug causes the underlying
         command to never receive a response if the board has no LEDs, requiring this
@@ -1153,45 +1238,36 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                     ]
                 )
 
-    async def _register_endpoint(
-        self,
-        endpoint,
-        profile_id=zigpy.profiles.zha.PROFILE_ID,
-        device_id=zigpy.profiles.zha.DeviceType.CONFIGURATION_TOOL,
-        device_version=0b0000,
-        latency_req=c.af.LatencyReq.NoLatencyReqs,
-        input_clusters=[],
-        output_clusters=[],
-    ):
+    async def _register_endpoints(self) -> None:
         """
-        Method to register an endpoint simultaneously with both zigpy and Z-Stack.
-
-        This lets us keep track of our own endpoint information without duplicating
-        that information again, and exposing it to higher layers (e.g. Home Assistant).
+        Registers the Zigbee endpoints required to communicate with various devices.
         """
 
-        # Create a corresponding endpoint on our Zigpy device first
-        zigpy_ep = self.zigpy_device.add_endpoint(endpoint)
-        zigpy_ep.profile_id = profile_id
-        zigpy_ep.device_type = device_id
-
-        for cluster in input_clusters:
-            zigpy_ep.add_input_cluster(cluster)
-
-        for cluster in output_clusters:
-            zigpy_ep.add_output_cluster(cluster)
-
-        zigpy_ep.status = zigpy.endpoint.Status.ZDO_INIT
-
-        return await self._znp.request(
+        await self._znp.request(
             c.AF.Register.Req(
-                Endpoint=endpoint,
-                ProfileId=profile_id,
-                DeviceId=device_id,
-                DeviceVersion=device_version,
-                LatencyReq=latency_req,  # completely ignored by Z-Stack
-                InputClusters=input_clusters,
-                OutputClusters=output_clusters,
+                Endpoint=1,
+                ProfileId=zigpy.profiles.zha.PROFILE_ID,
+                DeviceId=zigpy.profiles.zha.DeviceType.IAS_CONTROL,
+                DeviceVersion=0b0000,
+                LatencyReq=c.af.LatencyReq.NoLatencyReqs,
+                InputClusters=[clusters.general.Ota.cluster_id],
+                OutputClusters=[
+                    clusters.security.IasZone.cluster_id,
+                    clusters.security.IasWd.cluster_id,
+                ],
+            ),
+            RspStatus=t.Status.SUCCESS,
+        )
+
+        await self._znp.request(
+            c.AF.Register.Req(
+                Endpoint=2,
+                ProfileId=zigpy.profiles.zll.PROFILE_ID,
+                DeviceId=zigpy.profiles.zll.DeviceType.CONTROLLER,
+                DeviceVersion=0b0000,
+                LatencyReq=c.af.LatencyReq.NoLatencyReqs,
+                InputClusters=[],
+                OutputClusters=[],
             ),
             RspStatus=t.Status.SUCCESS,
         )
@@ -1409,6 +1485,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         return response
 
+    @combine_concurrent_calls
     async def _discover_route(self, nwk: t.NWK) -> None:
         """
         Instructs the coordinator to re-discover routes to the provided NWK.
@@ -1420,25 +1497,15 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if self._znp.version < 3.30:
             return
 
-        if nwk in self._route_discovery_futures:
-            return await self._route_discovery_futures[nwk]
+        await self._znp.request(
+            c.ZDO.ExtRouteDisc.Req(
+                Dst=nwk,
+                Options=c.zdo.RouteDiscoveryOptions.UNICAST,
+                Radius=30,
+            ),
+        )
 
-        future = asyncio.get_running_loop().create_future()
-        self._route_discovery_futures[nwk] = future
-
-        try:
-            await self._znp.request(
-                c.ZDO.ExtRouteDisc.Req(
-                    Dst=nwk,
-                    Options=c.zdo.RouteDiscoveryOptions.UNICAST,
-                    Radius=30,
-                ),
-            )
-
-            await asyncio.sleep(0.1 * 13)
-        finally:
-            future.set_result(True)
-            del self._route_discovery_futures[nwk]
+        await asyncio.sleep(0.1 * 13)
 
     async def _send_request(
         self,
@@ -1477,8 +1544,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         tried_assoc_remove = False
         tried_route_discovery = False
-        tried_disable_route_discovery_suppression = False
         tried_last_good_route = False
+        tried_ieee_address = False
 
         # Don't release the concurrency-limiting semaphore until we are done trying.
         # There is no point in allowing requests to take turns getting buffer errors.
@@ -1600,11 +1667,18 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                             # letting the retry mechanism deal with it simpler.
                             await self._discover_route(dst_addr.address)
                             tried_route_discovery = True
-                        elif not tried_disable_route_discovery_suppression:
-                            # Disable route discovery suppression. This appears to
-                            # generate a bit more network traffic.
-                            options &= ~c.af.TransmitOptions.SUPPRESS_ROUTE_DISC_NETWORK
-                            tried_disable_route_discovery_suppression = True
+                        elif (
+                            not tried_ieee_address
+                            and device is not None
+                            and dst_addr.mode == t.AddrMode.NWK
+                        ):
+                            # Try using the device's IEEE address instead of its NWK.
+                            # If it works, the NWK will be updated when relays arrive.
+                            tried_ieee_address = True
+                            dst_addr = t.AddrModeAddress(
+                                mode=t.AddrMode.IEEE,
+                                address=device.ieee,
+                            )
 
                         LOGGER.debug(
                             "Request failed (%s), retry attempt %s of %s",
